unit KModbus;

interface
{
   Component to do modbus protocol.

   Author:  Vesa Lappalainen
   Date:    28.12.1996
   Changes: 14.01.1997
             +
   Changes: 22.03.1997
             + handling for multiple slaves
             + Critical section

   Problems: - sometimes dosen't get the whole send string back?
             - no automatic sending to keep line alive
             - no handling for error's

   Can only following functions for Yaskawa inverter:
      3  = Read Holding registers
     16  = Preset Multiple Registers, can set only one or two at the time


     TModBus        -


}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  com485,StdCtrls, kErrors;

type
  TModbus = class(TComponent)
  private
    LastAnswer : ShortString;
    LastReg : integer;
    LastMode : integer;
    LastSlave : integer;
    Cri:TRTLCriticalSection; t:integer;

    procedure SetPort(nPortToSet: Integer);
    function GetPort:integer;
  protected
    procedure SetErrors(err:TErrors); virtual;

  public
    com:TCom485;
    FErrors : TErrors;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ErrorShowMessage(const s:String);
    procedure ShowDebug(const s:String);
    function CheckLen(const s:ShortString; n:integer) : boolean;
    function CheckCrc(const s:ShortString) : boolean;
    function CheckModbusError(const s:ShortString) : boolean;
    function CheckError(len:integer) : boolean;
    function Open: Boolean;
    function Close: Boolean;
    function IsOpen : Boolean;
    function SetReg(slave,r,v:integer):integer;
    function SetReg2(slave,r,v1,v2:integer):integer;
    function ReadReg(slave,r:integer):integer;
    function ReadReg2(slave,r:integer; var v1,v2:integer):integer;
    function Write(const s:ShortString;al:integer):string;
    function Read:String;  //Temp
  published
    property Port: Integer read GetPort write SetPort;
    property Errors : TErrors read FErrors write SetErrors;
  end;

type
  TModBusSlave = class(TObject)
  private
    fmodbus : TModbus;
    fslave : integer;
  protected
  public
    function SetReg(r,v:integer):integer;
    function SetReg2(r,v1,v2:integer):integer;
    function ReadReg(r:integer):integer;
    function ReadReg2(r:integer; var v1,v2:integer):integer;
    function Ok : boolean;
  published
    property modbus : TModbus read fmodbus write fmodbus;
    property slave : integer read fslave write fslave;
  end;


implementation

type TModbusMsg = record
       nr : integer;
       msg : String;
     end;

const KnownErrors = 4;
const ModbusErrors : Array[0..KnownErrors-1] of TModbusMsg = (
          ( nr:01 ; msg:'Illegal function'          ),
          ( nr:02 ; msg:'Illegal data address'      ),
          ( nr:33 ; msg:'Illegal value'             ),
          ( nr:34 ; msg:'Can not change during run' )
      );

const KnownFunctions = 2;
const ModbusFunctions : Array[0..KnownFunctions-1] of TModbusMsg = (
          ( nr:03 ; msg:'Read register'        ),
          ( nr:16 ; msg:'Write register'       )
      );

type TStr10 = String[10];

function ModBusMsg(n:integer; const Msgs : array of TModbusMsg):ShortString;
var i:integer;
begin
  for i:=0 to High(Msgs) do with Msgs[i] do
    if ( nr = n ) then begin Result := msg; Exit; end;
  Result := 'Unknown';
end;


function crc16_2(puchMsg:PChar;usDataLen:integer):word;
  stdcall; far; external;
{$l crc.obj}

function CountCRCInt(const s:Array of char; n:integer):integer;
var  crc : integer;
begin
  crc := crc16_2(s,n);
  Result := crc;
end;

{------------------------------------------------------------------------------}
function CharToHex(c:char):string;
begin
  Result := Format('%02x',[Ord(c)]);
  if ( Result[1] = ' ' ) then Result[1] := '0';
end;

function StrToHexStr(const s:ShortString):string;
var i:integer;
begin
  Result := '';
  for i:=1 to Length(s) do
    Result := Result + CharToHex(s[i]) + ' ';
end;


{------------------------------------------------------------------------------}
procedure TModbus.SetErrors(err:TErrors); 
begin
  FErrors := err;
  com.Errors := err;
end;

{------------------------------------------------------------------------------}
procedure TModbus.ErrorShowMessage(const s:String);
begin
  if ( Errors <> NIL ) and ( FErrors.ShowModbusError )  then
    Errors.Add(ceModbusError,'Modbus: '
                 + IntToStr(LastSlave)
                 + ' ' + ModbusMsg(LastMode,ModbusFunctions)
                 + '[' + IntToStr(LastReg) + '] '
                 + s
                 + ' : ' + StrToHexStr(LastAnswer)
               );
end;

{------------------------------------------------------------------------------}
procedure TModbus.ShowDebug(const s:String);
begin
  if ( Errors <> NIL ) and ( Errors.ShowModbusDebug )  then
    Errors.Add(ceModbusDebug,'Modbus: '
                 + IntToStr(LastSlave)
                 + ' ' + ModbusMsg(LastMode,ModbusFunctions)
                 + '[' + IntToStr(LastReg) + '] '
                 + s
                 + ' : ' + StrToHexStr(LastAnswer)
               );
end;

{------------------------------------------------------------------------------}
function TModbus.CheckLen(const s:ShortString; n:integer) : boolean;
begin
  Result := True;
  if ( Length(s) = n ) then exit;
  if ( Errors <> NIL )  and Errors.ShowModbusError   then
    ErrorShowMessage('Did not get ' + IntToStr(n) + ' bytes got only: ' +
                IntToStr(Length(s)));
  Result := false;
end;

{------------------------------------------------------------------------------}
function CrcStr(const s:ShortString):TStr10;
var crc : Integer;
begin
  crc := CountCRCInt(s[1],Length(s));
  Result[0] := char(2); // SetLength(Result,2);
  Result[1] := char(crc); Result[2] := char(hi(crc));
end;

{------------------------------------------------------------------------------}
function TModbus.CheckCrc(const s:ShortString) : boolean;
var crc : integer; crcs : string[4];
begin
  Result := True;
  If ( CountCRCInt(s[1],Length(s)) = 0 ) then exit;
  if ( Errors <> NIL )  and Errors.ShowModbusError then begin
    crc := CountCRCInt(s[1],Length(s)-2);
    crcs[0] := char(2); // SetLength(Result,2);
    crcs[1] := char(crc); crcs[2] := char(hi(crc));
    ErrorShowMessage('CRC error!'+ ' : ' + StrToHexStr(crcs));
  end;
  Result := false;
end;

{------------------------------------------------------------------------------}
function TModbus.CheckModbusError(const s:ShortString) : boolean;
var func : integer;
begin
  Result := True;
  if ( ord(s[2]) and $80 ) = 0  then exit;
  Result := false;
  if ( Not CheckCrc(s) ) then exit;
  func := ord(s[2]) and $7f;
  if ( Errors <> NIL )  and  Errors.ShowModbusError then
    ErrorShowMessage('Slave ' + IntToStr(ord(s[1]))
      + ' function ' + IntToStr(func)
      + ' (' + ModbusMsg(func,ModbusFunctions) + ')'
      + ' code: ' + IntToStr(ord(s[3]))
      + ' ('+ ModbusMsg(ord(s[3]),ModbusErrors) + ')'
      );

end;

{------------------------------------------------------------------------------}
function TModbus.CheckError(len:integer) : boolean;
begin
  Result := True;
  if ( not CheckModbusError(LastAnswer) ) then begin Result := True; Exit; end;
  if ( not CheckLen(LastAnswer,len) ) then begin Result := True; Exit;  end;
  if ( not CheckCrc(LastAnswer) ) then begin  Result := True; Exit;  end;
  if ( Ord(LastAnswer[1]) <> LastSlave ) then begin Result := True;
    ErrorShowMessage('wrong slave answer: '+IntToStr(Ord(LastAnswer[1])));
  end;
end;

{------------------------------------------------------------------------------}
{ Component constructor }
constructor TModbus.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  com := TCom485.Create(self);
  com.Port := 3;
  FErrors := NIL;
  InitializeCriticalSection(Cri);
end;

{------------------------------------------------------------------------------}
{ Component destructor }
destructor TModbus.Destroy;
begin
  DeleteCriticalSection(Cri);
  com.Free;             { close the com port (if open) }
  inherited Destroy;	{ destroy ancestor class }
end;

{------------------------------------------------------------------------------}
function TModbus.IsOpen: Boolean;
begin
  Result := com.IsOpen;
end;

{------------------------------------------------------------------------------}
function TModbus.Open: Boolean;
begin
  Result := com.Open;
end;

{------------------------------------------------------------------------------}
function TModbus.Close: Boolean;
begin
  com.Close;
  Result := true;
end;

{------------------------------------------------------------------------------}
procedure TModbus.SetPort(nPortToSet: Integer);
begin
  com.Port := nPortToSet;
end;

{------------------------------------------------------------------------------}
function TModbus.GetPort:integer;
begin
  Result := com.Port;
end;

function ShortStr(i:integer):TStr10; asm
  mov byte ptr [edx],2
  mov [edx+1],ah
  mov [edx+2],al
end;
{
function ShortStr(i:integer):TStr10;
begin
  Result[0] := char(2); // SetLength(Result,2);
  Result[1] := char(hi(i)); Result[2] := char(i);
end;
}

{------------------------------------------------------------------------------}
function TModbus.Write(const s:ShortString;al:integer):string;
begin
  com.Write(s+CrcStr(s));
  Result := com.ReadN(al);
end;

{------------------------------------------------------------------------------}
{
             id fkt 102H lkm=1 tav data  crc
   lhet :  01 10 01 02 00 01 02  00 06 37 70
             id fk osoite lkm   crc
   vastaus:  01 10 01 02 00 01 A1 F5
}
function TModbus.SetReg(slave,r,v:integer):integer;
var s:ShortString;
begin
  EnterCriticalSection(Cri);
  LastMode := 16;
  LastSlave := slave;
  LastReg := r;
  Result := 1;

  s := char(slave) + char(16) + ShortStr(r) + ShortStr(1) + char(2) + ShortStr(v);
  LastAnswer := Write(s,8);
  if ( not CheckError(8) ) then Result := 0;
  if ( Errors <> NIL ) and ( Errors.ShowModbusDebug )  then
    ShowDebug('SetReg: ' + IntToStr(v));
  LeaveCriticalSection(Cri);
end;

{------------------------------------------------------------------------------}
function TModbus.SetReg2(slave,r,v1,v2:integer):integer;
var s:ShortString;
begin
  EnterCriticalSection(Cri);
  LastMode := 16;
  LastSlave := slave;
  LastReg := r;

  Result := 0;
  s := char(slave) + char(16) + ShortStr(r) + ShortStr(2) + char(4) +
       ShortStr(v1) + ShortStr(v2);
  LastAnswer := Write(s,8);
  if ( not CheckError(8) ) then Result := 0;
  if ( Errors <> NIL ) and ( Errors.ShowModbusDebug )  then
    ShowDebug('SetReg2: ' + IntToStr(v1) + ' ' + IntToStr(v2));
  LeaveCriticalSection(Cri);
end;

{------------------------------------------------------------------------------}
{
                  id fk  rek  lkm    crc
        lhet :  01 03 01 0B 00 01 F4 34

                        rek  011
                        mem  10BH
                  id fkt #  230V  crc
        takaisin: 01 03 02 08 FC BF C5
                   1  2  3  4  5  6  7

    Jos funktio on vr, tulee takasin vastaus, jossa on
    id, 7-bit pll +fkt nro,  virhekoodi + crc

}
function TModbus.ReadReg(slave,r:integer):integer;
var s:ShortString;
begin
  EnterCriticalSection(Cri);
  LastMode := 3;
  LastSlave := slave;
  LastReg := r;

  s := char(slave) + char(3) + ShortStr(r) + ShortStr(1);
  LastAnswer := Write(s,7);

  Result := 256*Ord(LastAnswer[4]) + Ord(LastAnswer[5]);

  if ( not CheckError(7) ) then Result := 0;
  if ( Errors <> NIL ) and ( Errors.ShowModbusDebug )  then
    ShowDebug('ReadReg: ' + IntToStr(Result));
  LeaveCriticalSection(Cri);

end;

{------------------------------------------------------------------------------}
{
                  id fk  rek  lkm    crc
        lhet :  01 03 01 0B 00 02 ?? ??

                        rek  011   012
                        mem  10BH  10CH
                  id fkt #  230V  60 Hz,  crc
        takaisin: 01 03 04 08 FC  02 58  ?? ??
                   1  2  3  4  5  6  7   8   9

    Jos funktio on vr, tulee takasin vastaus, jossa on
    id, 7-bit pll +fkt nro,  virhekoodi + crc

}
function TModbus.ReadReg2(slave,r:integer; var v1,v2:integer):integer;
var s:ShortString;
begin
  EnterCriticalSection(Cri);
  LastMode := 3;
  LastSlave := slave;
  LastReg := r;

  Result := -1;
  s := char(slave) + char(3) + ShortStr(r) + ShortStr(2);
  s := Write(s,9);
  v1 := 256*Ord(s[4]) + Ord(s[5]);
  v2 := 256*Ord(s[6]) + Ord(s[7]);
  LastAnswer := s;

  if ( not CheckError(9) ) then Result := 0;
  if ( Errors <> NIL ) and ( Errors.ShowModbusDebug )  then
    ShowDebug('ReadReg2: ' + IntToStr(v1) + ' ' + IntToStr(v2));
  LeaveCriticalSection(Cri);
end;

function TModbus.Read:String;  //Temp
begin
  Result := StrToHexStr(LastAnswer); //com.Read;
end;

{------------------------------------------------------------------------------}
{ TModbusSlave ================================================================}
{------------------------------------------------------------------------------}

function TModbusSlave.SetReg(r,v:integer):integer;
begin
  Result := modbus.SetReg(slave,r,v);
end;

{------------------------------------------------------------------------------}
function TModbusSlave.SetReg2(r,v1,v2:integer):integer;
begin
  Result := modbus.SetReg2(slave,r,v1,v2);
end;

{------------------------------------------------------------------------------}
function TModbusSlave.ReadReg(r:integer):integer;
begin
  Result := modbus.ReadReg(slave,r);
end;

{------------------------------------------------------------------------------}
function TModbusSlave.ReadReg2(r:integer; var v1,v2:integer):integer;
begin
  Result := modbus.ReadReg2(slave,r,v1,v2);
end;

{------------------------------------------------------------------------------}
function TModbusSlave.Ok : boolean;
begin
  Result := (modbus <> nil) and ( modbus.IsOpen );
end;

end.
