unit KYasPar;
{
   Parameters for Yaskawa inverter

   Author:  Vesa Lappalainen
   Date:    31.12.1996
   Changes: 05.04.1997
              + reset ottaa kaikki RunParametrin biti pois plt aluksi
              (jos run-pll tai Ext.error, ei muuten resetoi).
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  IniPara,KModbus,kparams;


type
  TYasParamValue = class(TParamValue)
  private
    modbusslave:^TModBusSlave;

  protected
  public
    function WriteFunction(d: double): string; override;
    function ReadFunction: string; override;
  published
  end;

type
  TYasParams = class(TParams)
  private
    FYasForm : TForm;
    FModbusSlave:TModBusSlave;
    thz : Double;  // Temp for demos, where no modbus active

    RunParam,
    HzParam : TParamValue;

  protected
    procedure SetDirection(d:integer); virtual;
    function GetDirection:integer; virtual;
    function FindParams:boolean; virtual;
    procedure RefreshRunParams; virtual;
    procedure SendRunParams; virtual;
    function CreateList: string; override;

    function GetSlave:integer;  virtual;
    procedure SetSlave(s:integer); virtual;
    function GetModbus : TModbus;
    procedure SetModbus(mb:TModbus);
  public
    constructor Create(AOwner: TComponent); override;
    procedure WriteHz(hz:double); virtual;
    procedure WriteRunHz(hz:double); virtual;
    function ReadHz:double; virtual;
    function ReadMaxHz:double; virtual;
    function ReadMinHz:double; virtual;
        function WriteReadHz(hz:double):double; virtual;
    function AccAHz:double; virtual;
    function DecAHz:double; virtual;
    procedure Stop; virtual;
    procedure Reset;
    function NewParam(const name:string):TParamValue; override;
    procedure ShowControl; virtual;
    procedure ShowErrors; virtual;
  published
    property ModbusSlave:TModBusSlave read FModbusSlave;
    property Modbus:TModBus read GetModBus write SetModbus;
    property Direction:integer read GetDirection write SetDirection;
    property Slave : integer read GetSlave write SetSlave;
  end;

procedure Register;

implementation

uses YasForm;

const NameMaxHz   = 'n012';
      NameMinHz   = 'n017';
      NameAccTime = 'n019';
      NameDecTime = 'n020';

{------------------------------------------------------------------------------}
{ Write value to the fysical device if one exist and change value to d         }
{ if success                                                                   }
{ Return error message as text.                                                }
function TYasParamValue.WriteFunction(d: double): string;
begin
  value := d;
  Result := '';
  if ( not ModbusSlave.ok ) then exit;
  ModbusSlave.SetReg(Ind,AsInteger);
  AsInteger := ModbusSlave.ReadReg(Ind)
end;

{------------------------------------------------------------------------------}
{ Read value from fysical device and change the value if success               }
{ Return error message as text.                                                }
function TYasParamValue.ReadFunction: string;
begin
  Result := '';
  if ( not ModbusSlave.ok ) then exit;
  AsInteger := ModbusSlave.ReadReg(Ind)
end;

{------------------------------------------------------------------------------}
{ Overdriven virtual to create right params for this Params-list               }
function TYasParams.NewParam(const name:string):TParamValue;
var p : TYasParamValue;
begin
  p := TYasParamValue.Create;
  p.ModbusSlave := @FModbusSlave;
  Result := p;
end;

{------------------------------------------------------------------------------}
{ TYasParam -------------------------------------------------------------------}
{------------------------------------------------------------------------------}

{------------------------------------------------------------------------------}
function TYasParams.GetModbus : TModbus;
begin
  Result := ModbusSlave.Modbus;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.SetModbus(mb:TModbus);
begin
  ModbusSlave.Modbus := mb;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.SendRunParams;
var r,ihz : integer;
begin
  r   := RunParam.AsInteger;
  ihz := abs(HzParam.AsInteger);
  ModbusSlave.SetReg2(1,r,ihz);
end;

{------------------------------------------------------------------------------}
procedure TYasParams.RefreshRunParams;
var r,ihz : integer;
begin
  ModbusSlave.ReadReg2(1,r,ihz);
  RunParam.AsInteger := r;
  HzParam.AsInteger := ihz;
end;

{------------------------------------------------------------------------------}
function TYasParams.FindParams:boolean;
begin
  RunParam    := FindParam('r001');
  HzParam     := FindParam('r002');
  Result := (RunParam<>NIL) and (HzParam<>NIL);
end;

{------------------------------------------------------------------------------}
procedure TYasParams.SetDirection(d:integer);
begin
  RunParam.SetBit(1,d<0);
end;

{------------------------------------------------------------------------------}
function TYasParams.GetDirection:integer;
begin
  Result := 1;
  if ( RunParam.GetBit(1) )  then Result := -1;
end;

{------------------------------------------------------------------------------}
function TYasParams.ReadHz:double;
var hz:double;
begin
  Result := thz;
  if ( not ModbusSlave.Ok ) then exit;
  RefreshRunParams;
  hz := HzParam.Value;
  if ( direction = -1 ) then Hz := -Hz;
  Result := Hz;
  thz := hz;
end;

{------------------------------------------------------------------------------}
function TYasParams.ReadMaxHz:double;
begin
  if ( not ModbusSlave.Ok ) then Result := 60
  else Result := ReadValueByName(NameMaxHz);
end;

{------------------------------------------------------------------------------}
function TYasParams.ReadMinHz:double;
begin
  if ( not ModbusSlave.Ok ) then Result := 0
  else Result := ReadValueByName(NameMinHz);
end;

{------------------------------------------------------------------------------}
procedure TYasParams.WriteHz(hz:double);
begin
  thz := hz;
  if ( hz < 0 ) then direction := -1 else direction := 1;
  HzParam.Value := abs(hz);
  SendRunParams;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.WriteRunHz(hz:double);
begin
  RunParam.SetBit(0,abs(hz)>0.2);
  WriteHz(hz);
end;

{------------------------------------------------------------------------------}
function TYasParams.WriteReadHz(hz:double):double;
begin
  WriteRunHz(hz);
  Result := ReadHz;
end;

{------------------------------------------------------------------------------}
function TYasParams.AccAHz:double;
var t : double;
begin
  t := ReadValueByName(NameAccTime);
  if ( t = 0 ) then t := 1;
  Result := ReadMaxHz/t;
end;

{------------------------------------------------------------------------------}
function TYasParams.DecAHz:double;
var t : double;
begin
  t := ReadValueByName(NameDecTime);
  if ( t = 0 ) then t := 1;
  Result := ReadMaxHz/t;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.Stop;
begin
  RunParam.WriteBit(0,false);
end;

{------------------------------------------------------------------------------}
procedure TYasParams.Reset;
begin
  RunParam.WriteValue(0);
  RunParam.WriteBit(3,true);
  RunParam.WriteBit(3,false);
end;

{------------------------------------------------------------------------------}
function TYasParams.CreateList: string;
begin
  Result := Inherited CreateList;
  if ( Result = '' ) and ( Not FindParams ) then Result := 'Param error';
end;

{------------------------------------------------------------------------------}
constructor TYasParams.Create(AOwner: TComponent);
begin
  FYasForm := NIL;
  FmodbusSlave := TModbusSlave.Create;
  FModbusSlave.slave := 1;
  thz := 0;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.ShowControl;
begin
  if ( FYasForm = NIL ) then
    FYasForm := TFormYas.Create2(@FYasForm,Self);
  if ( FYasForm <> NIL ) then FYasForm.Show;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.ShowErrors;
begin
  Modbus.Errors.ShowErrors;
end;

{------------------------------------------------------------------------------}
function TYasParams.GetSlave:integer;
begin
  Result := ModbusSlave.Slave;
end;

{------------------------------------------------------------------------------}
procedure TYasParams.SetSlave(s:integer);
begin
  ModbusSlave.Slave := s;
end;



procedure Register;
begin
end;

end.
